import typer
from pathlib import Path
import xml.etree.ElementTree as ET
from xml.dom import minidom

app = typer.Typer()


def make_dynamic(element, key):
    _id = element.get("id")
    if key is None:
        value = element.text
        s = f"{_id}={value}"
        element.text = f"$({s})"
    else:
        value = element.get(key)
        if "$" in value:
            # Don't replace defaults that are already set
            # value = value.replace("$", "\$")
            return None
        if "=" in value:
            # https://github.com/CompEvol/beast2/blob/4d95f0095d20a6999d2757299c629bfbbf14b94b/src/beast/util/XMLParser.java#L372
            # the way that defaults are passed means you can't use '=' in there values...
            # https://github.com/CompEvol/beast2/pull/991
            return None
        s = f"{_id}.{key}={value}"
        element.set(key, f"$({s})")


def make_all_dynamic(element):
    for key in filter(lambda k: k != "id", element.keys()):
        make_dynamic(element, key)
    if element.tag == "parameter":
        make_dynamic(element, None)


def add_mc3_options(run):
    # <run id="mcmc" spec="beast.coupledMCMC.CoupledMCMC" chainLength="10000000" chains="4" target="0.234" logHeatedChains="true" deltaTemperature="0.1" optimise="true" resampleEvery="1000" >
    mc3_options = {
        "spec": "beast.coupledMCMC.CoupledMCMC",
        "chains": "2",
        "target": "0.234",
        "logHeatedChains": "false",
        "resampleEvery": "100",
        "tempDir": "",
        "deltaTemperature": "0.1",
        # 'maxTemperature': '',  # cannot be set to ''
        "optimise": "true",
        "optimiseDelay": "100",
        "preSchedule": "true",
    }
    for option in mc3_options:
        run.set(option, mc3_options[option])


def add_ps_options(root):
    # <run spec='beast.inference.PathSampler' chainLength="100000" alpha='0.3' rootdir='/tmp/step' burnInPercentage='80' preBurnin="50000" deleteOldLogs='true'>
    ps_options = {
        "id": "ps",
        "nrOfSteps": "8",
        "spec": "beast.inference.PathSampler",
        "chainLength": "100000",
        "alpha": "0.3",
        "rootdir": "/tmp",
        "burnInPercentage": "50",
        "preBurnin": "100000",
        "deleteOldLogs": "false",
        "doNotRun": "false",
    }
    ps = ET.SubElement(root, "run", ps_options)
    ps.text = "\ncd $(dir)\njava -cp $(java.class.path) beast.app.beastapp.BeastMain $(resume/overwrite) -java -seed $(seed) beast.xml\n"
    return ps


def add_ns_options(run):
    # <run id="mcmc" spec="beast.gss.NS" chainLength="20000" particleCount="1" subChainLength="5000" epsilon="1e-12">
    ns_options = {
        "id": "mcmc",
        "spec": "beast.gss.MultiThreadedNS",
        "threads": "4",
        "chainLength": "20000",
        "particleCount": "1",
        "subChainLength": "5000",
        "epsilon": "1e-12",
    }
    for option in ns_options:
        run.set(option, ns_options[option])


def apply_optimise(path_to_output: Path, run):
    ops = []
    with open(path_to_output) as output:
        # extract optimization lines
        for line in output:
            if line.startswith("Operator"):
                break
        for line in output:
            if line == "\n":
                break
            ops.append(line)
    for l in ops:
        if "Try" not in l:
            continue
        ops_values = l.split()
        op_id = ops_values[0].split("(")[1][:-1]
        key = ops_values[-4]
        value = ops_values[-1]
        op_els = run.findall(f".//*[@id='{op_id}']")
        if not op_els:
            raise ValueError(f"Operator with 'id' {op_id} not found!")
        op_el = op_els[0]
        s = f"{op_id}.{key}={value}"
        op_el.set(key, f"$({s})")


def create_dynamic_xml(
    beast_xml, outfile=None, mc3=False, ps=False, ns=False, optimise=None
):
    tree = ET.parse(beast_xml)
    root = tree.getroot()
    run = root.find("run")
    if mc3:
        add_mc3_options(run)
    if ps:
        root.remove(run)
        ps = add_ps_options(root)
        run.tag = "mcmc"
        ps.append(run)
        run = ps
    if ns:
        add_ns_options(run)

    for el in run.iter():
        if "idref" in el.keys():
            continue
        make_all_dynamic(el)

    if optimise:
        apply_optimise(optimise, run)

    if outfile:
        tree.write(outfile)
    else:
        xmlstr = minidom.parseString(ET.tostring(root)).toprettyxml(newl="")
        typer.echo(xmlstr)


@app.command()
def main(
    beast_xml: Path,
    outfile: Path = typer.Option(None, help="Path to save the dynamic BEAST XML file."),
    mc3: bool = typer.Option(False, help="Add default MC3 options to XML file."),
    ps: bool = typer.Option(False, help="Add default PathSampler options to XML file."),
    ns: bool = typer.Option(
        False, help="Add default Multi threaded nested sampling options to XML file."
    ),
    optimise: Path = typer.Option(
        None,
        help="Path to previous run output file. Operator optimisations will be extracted.",
    ),
):
    """
    Dynamic BEAST XML
    """
    create_dynamic_xml(beast_xml, outfile, mc3, ps, ns, optimise)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dynamic_beast']

package_data = \
{'': ['*']}

install_requires = \
['typer[all]>=0.3.2']

entry_points = \
{'console_scripts': ['dynamic-beast = dynamic_beast.main:app']}

setup_kwargs = {
    'name': 'dynamic-beast',
    'version': '1.7.5',
    'description': '',
    'long_description': '# Dynamic BEAST\n\n[![PyPi](https://img.shields.io/pypi/v/dynamic-beast.svg)](https://pypi.org/project/dynamic-beast/)\n[![tests](https://github.com/Wytamma/dynamic-beast/actions/workflows/test.yml/badge.svg)](https://github.com/Wytamma/dynamic-beast/actions/workflows/test.yml)\n[![cov](https://codecov.io/gh/Wytamma/dynamic-beast/branch/master/graph/badge.svg)](https://codecov.io/gh/Wytamma/dynamic-beast)\n\nThis command line tool can be used to create a dynamic version of [BEAST2](http://www.beast2.org/) XML files. This dynamic XML file can be used to set BEAST parameters at runtime, which can be useful for testing different configurations or quickly modifying parameters without having to edit the XML file. \n\n## Install\nInstall `dynamic-beast` with pip (requires python -V >= 3.6.2).\n\n```\npip install dynamic-beast\n```\n\n## Usage\n\nGive `dynamic-beast` the path to a BEAST2 XML file and specify where to save the dynamic XML file (if `--outfile` is not specified XML will be printed to stdout).\n\n```bash\ndynamic-beast hcv_coal.xml > dynamic_hcv_coal.xml\n```\n\nThis will produce a `dynamic_hcv_coal.xml` file that can be used as standard in a BEAST analysis.\n\n```bash\nbeast dynamic_hcv_coal.xml\n```\n\nTo modify parameters at runtime use the `beast` definitions option `-D`.\n\n```bash\n# Change the chain length to 1000. \nbeast -D \'mcmc.chainLength=1000\' dynamic_hcv_coal.xml\n``` \n\nMultiple definitions can be passed at the same time.\n\n```bash\n# Change the treelog and tracelog sampling freq to 10000. \nbeast -D \'treelog.logEvery=10000,tracelog.logEvery=10000\' dynamic_hcv_coal.xml\n``` \n\nThe full `id` of a parameter you\'d like to set must be specified. \n\n```bash \nbeast -D \'clockRate.c:hcv=7.9E-4\' dynamic_hcv_coal.xml\n```\n\n## Explanation\n\nThe `dynamic-beast` tool replaces all the parameter values in the XML file with the `$(id.key=value)` format. The value variable is the default value that was initially specified in the XML file. However, the value can be redefined when running a BEAST analysis by making use of the [BEAST2 definitions option](https://www.beast2.org/2021/03/31/command-line-options.html#-d) (`-D`) that allows for user specified values. \n\nTo ensure reproducibility you should recreate static XML files of runs using dynamic parameters, this can be achieved using the `-DFout` argument e.g., `beast -D \'clockRate.c:hcv=7.9E-4\' -DFout static_hcv_coal.xml dynamic_hcv_coal.xml`. \n\n## Addtional features \n\n### CoupledMCMC\n\nMC3 options for the BEAST package [CoupledMCMC](https://github.com/nicfel/CoupledMCMC) can be added by using the `--mc3` option. This will add the default CoupledMCMC options which can then be configured at runtime with `-D`. \n\n```bash\n# Create dynamic MC3 XML \ndynamic-beast --mc3 hcv_coal.xml > dynamic_mc3_hcv_coal.xml \n# Configure MC3 with BEAST\nbeast -D \'mcmc.chains=4\' dynamic_mc3_hcv_coal.xml\n```\n\n### Path Sampling (Stepping Stone)\n\nPath sampling options for the package [model-selection](https://github.com/BEAST2-Dev/model-selection) can be add by using the `--ps` option. This will add the default model-selection options (e.g. stepping stone) which can then be configured at runtime with `-D`. \n\n```bash\n# Create dynamic Path Sampling XML \ndynamic-beast --ps hcv_coal.xml > dynamic_ps_hcv_coal.xml\n# Configure Path Sampling with BEAST\nbeast -D "ps.doNotRun=true,ps.rootdir=$(pwd)" dynamic_ps_hcv_coal.xml\n```\n\n### Multi threaded nested sampling\n\nMulti threaded nested sampling for the package [nested-sampling\n](https://github.com/BEAST2-Dev/nested-sampling) can be add by using the `--ns` option. This will add the default model-selection options which can then be configured at runtime with `-D`. \n\n```bash\n# Create dynamic Nested Sampling XML \ndynamic-beast --ns hcv_coal.xml > dynamic_ns_hcv_coal.xml\n# Configure Path Sampling with BEAST\nbeast -D "mcmc.threads=6,mcmc.chainLength=40000" dynamic_ns_hcv_coal.xml\n```\n\n### Auto apply optimisation suggestion\n\nAt the end of a analysis BEAST provides suggestions for optimising operators e.g. `Try setting scaleFactor to about 0.96`. See the end of the [example file](https://github.com/Wytamma/dynamic-beast/blob/master/data/Heterochronous_H3N2.out#L5366). A path to the output file can be provided to the `--optimise` flag and the suggestions will automatically be extracted and applied to the generated dynamic XML file. \n\n```bash\ndynamic-beast --optimise hcv_coal.out hcv_coal.xml > dynamic_hcv_coal.xml\n```\n\nDynamic-beast will look for a line starting with `Operator` and extract the suggestion from the lines that follow. So if you make your own `.out` file (i.e. by copy-pasting the BEAST output) you need to make sure the file starts with `Operator` on the first line. \n',
    'author': 'Wytamma Wirth',
    'author_email': 'wytamma.wirth@me.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
